import numpy as np
import pytest
from astropy.table.column import Column

from kanon.tables.hcolumn import HColumn
from kanon.units import Historical, Sexagesimal


class TestHColumn:
    def test_init(self):

        col = HColumn([1, 2, 3])

        assert col.significant is None

        with pytest.raises(AttributeError):
            assert col.ceil()

        bcol = col.astype(Sexagesimal)

        with pytest.raises(ValueError) as err:
            bcol[1] = 3

        assert "Sexagesimal" in str(err.value)

        assert "Sexagesimal" in repr(bcol)
        assert repr(col)[2:] == repr(Column(col))[1:]

    def test_astype(self):
        col = HColumn([1, 2, 3])

        bcol = col.astype(Sexagesimal)

        assert bcol.basedtype == Sexagesimal
        assert not bcol.attrs_equal(col)

        reconverted = bcol.astype(int)
        assert reconverted.attrs_equal(col)
        assert reconverted.basedtype is None

        assert np.array_equal(col.astype(float).astype(Sexagesimal), bcol)
        assert np.array_equal(bcol.astype(Historical), bcol)

        with pytest.raises(ValueError):
            strcol = HColumn(["a", "b", "c"])
            strcol.astype(Sexagesimal)

    def test_truncable(self):
        bcol = HColumn([1, 2, 3]).astype(Sexagesimal)

        assert bcol.significant == 0

        bcol = bcol.resize(2)
        bcol += Sexagesimal(31) >> 2

        assert bcol.significant == 2

        assert np.array_equal(
            [
                Sexagesimal("1;1"),
                Sexagesimal("2;1"),
                Sexagesimal("3;1"),
            ],
            bcol.ceil(1),
        )
        assert np.array_equal([1, 2, 3], bcol.truncate(0))
        assert np.array_equal(
            [
                Sexagesimal("1;0"),
                Sexagesimal("2;0"),
                Sexagesimal("3;0"),
            ],
            bcol.floor(1),
        )
        assert np.array_equal(
            [
                Sexagesimal("1;1"),
                Sexagesimal("2;1"),
                Sexagesimal("3;1"),
            ],
            round(bcol, 1),
        )



"""
The `~kanon.calendars` package provides everything you need to manipulate historical dates
with ancient calendars. All calculations on dates are based on Julian Day Numbers, starting
the 1st January 4713 B.C. at 12:00. This module already has a few of the major historical
calendars subclassed (see `calendars`). Once subclassed, a `Calendar` is created with an
`~kanon.calendars.calendars.Era` to define its beginning.

You can find all the calendars already instantiated in the `Calendar.registry` :

+---------------------+-----------------------+------------------------+------------------------+-------------------------------+
| `~calendars.Arabic` | `~calendars.Egyptian` | `~calendars.Julian`    | `~calendars.Byzantine` | `~calendars.Persian`          |
+---------------------+-----------------------+------------------------+------------------------+-------------------------------+
| Civil Hijra         | Nabonassar            | A.D.                   | A.D.                   | Yazdigird Andarjah at the end |
+---------------------+-----------------------+------------------------+------------------------+-------------------------------+
| Astronomical Hijra  | Philippus             | A.D. Leap December     |                        | Yazdigird Andarjah after Ābān |
+---------------------+-----------------------+------------------------+------------------------+-------------------------------+
|                     |                       | A.D. First month March |                        |                               |
+---------------------+-----------------------+------------------------+------------------------+-------------------------------+
|                     |                       | Julian Era             |                        |                               |
+---------------------+-----------------------+------------------------+------------------------+-------------------------------+

You can easily convert dates between all kinds of calendars. For this we use `Date` objects.
To start, let's try to build a `Date` of the 10th March 1324 from the Julian A.D. calendar :

>>> julian_ad = Calendar.registry["Julian A.D."]
>>> date = Date(julian_ad, (1324, 3, 10))
>>> str(date)
'10 Martius 1324 A.D. in Julian 12:00'
>>> date.jdn
2204718.0

Then we can convert this date in the Arabic Civil Hijra calendar :

>>> arabic_civil_hijra = Calendar.registry["Arabic Civil Hijra"]
>>> arabic_date = date.to_calendar(arabic_civil_hijra)
>>> str(arabic_date)
'13 Rabīʿ al-awwal 724 Civil Hijra in Arabic 12:00'
>>> arabic_date.jdn
2204718.0

We have succesfully converted a date expressed in one calendar into another. And we see that its
absolute date value (expressed in Julian Day Numbers) stays the same.
"""
from .calendars import Calendar, Date

__all__ = ("Calendar", "Date")



import glob
from typing import Any, Dict, List

import nbformat
import pytest
from astropy.units import degree
from nbformat import NotebookNode
from papermill.execute import execute_notebook

from kanon.units import Sexagesimal


def test_no_output():
    nbs: List[NotebookNode] = [
        nbformat.read(nb, nbformat.NO_CONVERT) for nb in glob.glob("./examples/*ipynb")
    ]
    for nb in nbs:
        for cell in nb.cells:
            assert "outputs" not in cell or len(cell.outputs) == 0


def get_nb(name: str, params: Dict[str, Any]) -> NotebookNode:
    nb = execute_notebook(f"./examples/{name}.ipynb", "-", parameters=params)
    return nb


@pytest.mark.parametrize(
    "params,result",
    [
        ({"year": 1327, "month": 7, "day": 3}, "1,47;18,48"),
        ({"year": 1691, "month": 9, "day": 9}, "2,55;31,33"),
        ({"year": 998, "month": 10, "day": 2}, "03,13;25,33"),
        ({"year": 1998, "month": 2, "day": 10}, "05,34;29,06"),
    ],
)
def test_sun_true_position(params, result):
    data = get_nb("sun_true_position", params).cells[-1].outputs[0].data["text/latex"]
    assert data == (Sexagesimal(result) * degree)._repr_latex_()


@pytest.mark.parametrize(
    "params,result",
    [
        ({"OBLIQUITY": "23;51,20"}, ("02;01,12,38", "23;51,20,22")),
        ({"OBLIQUITY": "0"}, ("0", "0")),
    ],
)
def test_declination(params, result):
    data: str = get_nb("declination", params).cells[7].outputs[0].data["text/html"]
    lines = data.split("<tr>")
    line5 = [li for li in lines if "<td>05 ;" in li][0]
    line90 = [li for li in lines if "<td>01,30 ;" in li][0]
    assert repr(Sexagesimal(result[0])).strip() in line5
    assert repr(Sexagesimal(result[1])).strip() in line90


@pytest.mark.parametrize(
    "params,result",
    [
        ({"OBLIQUITY": "23;51,20"}, ("0.3111", "-3.0222")),
        ({"OBLIQUITY": "0"}, ("54610.2778", "55083.2889")),
    ],
)
def test_ptolemy_viz(params, result):
    output = "".join(o["text"] for o in get_nb("ptolemy_viz", params).cells[9].outputs)
    assert f"mean : {result[0]}" in output
    assert f"mean : {result[1]}" in output



import math

import hypothesis.strategies as st
import pytest
from hypothesis import given

from kanon.calendars import Calendar, Date
from kanon.calendars.calendars import (
    Era,
    Julian,
    float_to_hm,
    hm_to_float,
    hours_to_day,
)
from kanon.units import Sexagesimal


def test_init():
    era = Era("test", 10)

    assert era.days_from_epoch(20) == 10

    new_cal = Julian(era)
    assert new_cal.era == era
    assert new_cal.name in Calendar.registry

    assert new_cal.__repr__() == "Calendar(Julian test)"

    with pytest.raises(ValueError):
        Julian(era)

    with pytest.raises(ValueError):
        new_cal.jdn_at_ymd(1, 50, 1)
    with pytest.raises(ValueError):
        new_cal.jdn_at_ymd(1, 1, 50)
    with pytest.raises(ValueError):
        new_cal.jdn_at_ymd(0, 1, 10)


def test_first_day():
    for cal in Calendar.registry.values():
        assert cal.jdn_at_ymd(1, 1, 1) == cal.era.epoch
        assert Date(cal, (1, 1, 1)).days_from_epoch() == 0


def test_from_julian():
    cal = Calendar.registry["Julian A.D."]
    days = cal.jdn_at_ymd(2, 1, 1)

    assert cal.from_julian_days(days).jdn == days
    assert cal.from_julian_days(640234).jdn == 640234
    assert cal.from_julian_days(cal.era.epoch - 1).jdn == cal.era.epoch - 1


@given(st.integers(1, 10000), st.integers(1, 12), st.integers(1, 28))
def test_dates_hypothesis(y: int, m: int, d: int):
    cal = Calendar.registry["Julian A.D."]

    date = Date(cal, (y, m, d))

    assert cal.from_julian_days(date.jdn) == date
    assert cal.get_time(y, m, d) == date.to_time()


def test_conversion():
    arabic = Calendar.registry["Arabic Civil Hijra"]
    julian = Calendar.registry["Julian A.D."]

    date = Date(arabic, (1, 1, 1))
    conversion = date.to_calendar(julian)
    assert conversion.ymd == (622, 7, 16)


def test_date():
    cal = Calendar.registry["Julian A.D."]
    date = Date(cal, (-4713, 1, 1))

    assert date.to_time().to_value("jd") == 0
    assert str(date) == "1 Ianuarius -4713 A.D. in Julian 12:00"

    date = Date(cal, (-1, 1, 1))

    assert date.to_time().to_value("jd") == 1721058
    assert str(date) == "1 Ianuarius -1 A.D. in Julian 12:00"

    date = Date(cal, (20, 3, 12))
    assert (date + 1).jdn == date.jdn + 1
    assert (date - 1).jdn == date.jdn - 1

    assert "13:30" in str(Date(cal, (1, 1, 1), Sexagesimal("13; 30")))

    with pytest.raises(ValueError) as err:
        Date(cal, (1, 1, 1), -1)

    assert "Time must be" in str(err)

    assert Date(cal, (2021, 7, 1), 0.2) + 0 == Date(cal, (2021, 7, 1), 0.2)


def test_jdn_at_ymd():
    cal = Calendar.registry["Julian A.D."]

    assert cal.jdn_at_ymd(-1, 12, 31) + 1 == cal.jdn_at_ymd(1, 1, 1)
    assert cal.jdn_at_ymd(-2, 12, 31) + 1 == cal.jdn_at_ymd(-1, 1, 1)

    assert cal.jdn_at_ymd(61, 1, 1) == 1743339

    assert cal.jdn_at_ymd(60, 12, 31) != cal.jdn_at_ymd(61, 1, 1)


def test_persian():
    cal_normal = Calendar.registry["Persian Yazdigird Andarjah at the end"]
    cal_variant = Calendar.registry["Persian Yazdigird Andarjah after Ābān"]

    assert len(cal_normal.months) == len(cal_variant.months)
    assert cal_normal.months[-1].name == cal_variant.months[8].name


def test_dates_fraction():
    cal = Calendar.registry["Julian A.D."]

    assert cal.from_julian_days(5000.6) != cal.from_julian_days(5000)

    date = Date(cal, (5, 2, 3), Sexagesimal(24) * 0.4)

    assert math.isclose(hours_to_day((date + 1).hours), 0.4)

    assert (date + 1).ymd == (5, 2, 4)


def test_time_convert():
    assert float_to_hm(hm_to_float(12, 30)) == (12, 30)

    with pytest.raises(ValueError) as err:
        hm_to_float(-1, 5)

    assert "Incorrect" in str(err)

    with pytest.raises(ValueError) as err:
        float_to_hm(2)

    assert "Incorrect" in str(err)



import pytest
from astropy.units import Quantity
from astropy.units import arcminute as _arcminute
from astropy.units import degree as _degree
from astropy.units.core import Unit

from kanon.units import Sexagesimal
from kanon.units.radices import BasedQuantity

degree: Unit = _degree
arcminute: Unit = _arcminute


class TestQuantity:
    def test_init_basedquantity(self):
        q = Sexagesimal(1) * degree
        assert isinstance(q, BasedQuantity)
        assert q.unit == degree
        assert q.value.equals(Sexagesimal(1))
        q = Sexagesimal(1) / degree
        assert isinstance(q, BasedQuantity)
        assert q.unit == 1 / degree
        assert q.value.equals(Sexagesimal(1))

        assert type(BasedQuantity(1, degree)) is Quantity

    def test_attribute_forwarding(self):
        q = Sexagesimal("1;0,1,31") * degree

        with pytest.raises(AttributeError):
            q._from_string()
        with pytest.raises(AttributeError):
            q.does_not_exist()

        assert q.truncate(2).value.equals(q.value.truncate(2))
        assert q.left == (1,)

        assert round(q, 2).value.equals(round(q.value, 2))

    @pytest.mark.filterwarnings("ignore")
    def test_shifting(self):
        q = Sexagesimal("1;0,1,31") * degree

        assert (q << 2).value.equals(q.value << 2)
        assert (q >> 2).value.equals(q.value >> 2)

        arcmq = q << 1 * arcminute
        assert arcmq.unit == arcminute
        assert arcmq.value.equals(Sexagesimal("1,0;1,31,0"))

        with pytest.raises(TypeError):
            q >> 1 * arcminute



import math as m

from . import utils
from .meta import dmodel
from .table_types import (
    Jupiter,
    Mars,
    Mercury,
    Moon,
    SphericalAstronomical,
    Sun,
    Mathematical,
    Venus,
)
from .utils import DEG, RAD

# # # # # # # # # # #
#        SUN        #
# # # # # # # # # # #


@dmodel(Sun.equ_of_the_sun, 23, 50)
def equ_of_the_sun(x: float, e: float) -> float:
    """
    Sun equation
    :param x:
    :param: solar eccentricity
    :return:
    """
    return DEG * m.atan(utils.product_sine_0(x, e) / (60 + e * m.cos(x * RAD)))


@dmodel(Sun.mean_motion_solar_tropical_long, 34, 34)
def long_of_the_tropical_mean_sun(x, val):
    """
    Longitude of the  tropical mean sun
    :param x: value in day
    :param val: value for 1 day
    :return:
    """
    return val * x


@dmodel(Sun.equ_of_time, 51, 25, 26, 27, 28, 29)
def equ_of_times_for_true_sun(x, obl, e, ap, epo, hdeg):
    """
    Equation of times for true sun
    :param x: true longitude of the sun in degree
    :param obl: obliquity of the ecliptic in degree
    :param e: solar eccentricity
    :param ap: solar apogee longitude in degree
    :param epo: epoch constant in degree
    :param hdeg: hour conversion rate, deg->h
    :return:
    """

    q = -DEG * m.asin(e * m.sin(RAD * (x - ap)) / 60)

    return hdeg * (x - q - utils.right_asc_0(x, obl) + epo)


@dmodel(Sun.equ_of_time, 52, 25, 26, 27, 28, 29)
def equ_of_times_for_mean_sun(x, obl, e, ap, epo, hdeg):
    """
    Equation of times for mean sun
    :param x: mean longitude in degree
    :param obl: obliquity of the ecliptic
    :param e: solar eccentricity
    :param ap: solar apogee longitude
    :param epo: epoch constant
    :param hdeg: hour conversion rate: deg ->h
    :return: equation of time in hour
    """
    q = utils.q_0(x - ap, e)

    return hdeg * (x + epo - utils.right_asc_0(x + q, obl))


# # # # # # # # # # #
#      MERCURY      #
# # # # # # # # # # #


@dmodel(Mercury.equ_anomaly_at_mean_dist, 32, 104)
def equ_of_anomaly_mercury_at_mean_dist(x, R):
    """
    Equation of anomaly Mercury at mean distance
    :param x: true argument in degree
    :param R: radius of the epicycle
    :return: mercury anomaly equation in degree
    """
    return utils.planet_anomaly_0(x, R, 60)


@dmodel(Mercury.equ_center, 36, 101)
def equ_of_center_of_mercury(x, e):
    """
    Equation of center of Mercury
    :param x: mean centre in degree
    :param e: eccentricity
    :return:  centre equation in degree (negative on [0,180º])
    """
    s = m.sqrt(60 ** 2 - (e * (m.sin(x * RAD) + m.sin(2 * x * RAD))) ** 2) + e * (
        m.cos(x * RAD) + m.cos(2 * x * RAD)
    )

    return -DEG * m.atan(utils.product_sine_0(x, e) / (s + e * m.cos(x * RAD)))


@dmodel(Mercury.total_equ_double_arg_table, 53, 124, 125)
def planet_double_arg_mercury(x, y, e, R):
    """
    Planet double argument mercury
    :param x: mean center in degree
    :param y: true argument in degree
    :param e: eccentricity
    :param R: radius of the epicycle
    :return:
    """

    s = utils.s_m_0(x, e)
    rh = utils.rho_0(x, s, e)

    p = utils.planet_anomaly_0(y, R, rh)
    q = utils.q_1(x, e)

    return q + p


@dmodel(Mercury.equ_anomaly_at_max_dist, 39, 106, 107)
def equ_of_anomaly_mercury_at_great_dist(x, e, R):
    """
    Equation of anomaly Mercury at great distance
    :param x:
    :param e: eccentricity
    :param R: radius of the epicycle
    :return:
    """
    return utils.planet_anomaly_0(x, R, 60 + 3 * e)


@dmodel(Mercury.equ_anomaly_at_min_dist, 40, 109, 110)
def equ_of_anomaly_mercury_at_near_dist(x, e, R):
    """
    Equation of anomaly Mercury at near distance
    :param x:
    :param e: eccentricity
    :param R: radius of the epicycle
    :return:
    """
    rho = m.sqrt(60 ** 2 - 180 * e + 3 * e ** 2)
    return utils.planet_anomaly_0(x, R, rho)


@dmodel(Mercury.equ_minuta_proportionalia, 63, 221, 488)
def mercury_equ_proportional_minutes(x, y, e, R):
    """
    proportional minutes giving the approximate value of the equation of anomaly
    p(av,cm)in the Mercury case
    :param x: true argument in degree
    :param y: mean center in degree
    :param e: eccentricity
    :param R: radius of the epicycle
    :return:
    """

    p0 = utils.planet_anomaly_0(x, R, 60)
    p1 = utils.planet_anomaly_0(x, R, 60 + 3 * e)
    p2 = utils.planet_anomaly_0(x, R, utils.rho_2(120, e))
    maxp0 = utils.max_equat_anomaly_0(R, 60)
    maxp1 = utils.max_equat_anomaly_0(R, 60 + 3 * e)
    maxp2 = utils.max_equat_anomaly_0(R, utils.rho_2(120, e))
    maxp = utils.max_equat_anomaly_0(R, utils.rho_2(y, e))
    f1 = (maxp - maxp0) / (maxp0 - maxp1)
    f2 = (maxp - maxp0) / (maxp2 - maxp0)
    cm0 = utils.mean_pos_epic_center_1(e)
    cm = y
    if y > 180:
        cm = 360 - y
    if 0 <= cm <= cm0:
        p = p0 + (p0 - p1) * f1
    else:
        p = p0 + (p2 - p0) * f2
    return p


@dmodel(Mercury.planetary_stations, 64, 182, 183, 184, 185)
def first_stationary_point_of_mercury_by_proportional_minutes(x, s0, s1, s2, e):
    """
    First stationary point of Mercury by proportional minutes
    :param x: mean centre in degree (0<=x<360)
    :param s0: first station true argument at the apogee in degree (cm=0)
    :param s1: first station true argument at the mean position in degree (cm=cm0)
    :param s2: first station true argument at the perigee in degree(cm=120)
    :param e: planetary eccentricity
    :return: true argument of the mercury first station in degree according x
    """
    cm0 = utils.mean_pos_epic_center_1(e)
    cm = x
    if (180 < x) and (x < 360):
        cm = 360 - x
    R0 = utils.rho_2(0, e)
    R1 = utils.rho_2(cm0, e)
    R2 = utils.rho_2(120, e)
    R = utils.rho_2(cm, e)
    if 0 <= cm <= cm0:
        s = s0 + (s1 - s0) * (R - R0) / (R1 - R0)
    else:
        s = s1 + (s2 - s1) * (R - R1) / (R2 - R1)
    return s


@dmodel(Mercury.planetary_stations, 65, 185, 186, 187)
def first_stationary_point_of_mercury_by_calculation_and_proportional_minutes(x, e, R, vq):
    """
    First stationary point of Mercury by calculation and proportional minutes
    :param x: mean centre in degree (0<=x<360)
    :param e: planetary eccentricity
    :param R: planetary epicycle radius
    :param vq: velocity quotient
    :return:true argument of the mercury first station in degree
    """

    s0 = utils.first_stationary_point_mercury_0(0, vq, R, e)
    cm0 = utils.mean_pos_epic_center_1(e)
    s1 = utils.first_stationary_point_mercury_0(cm0, vq, R, e)
    s2 = utils.first_stationary_point_mercury_0(120, vq, R, e)
    R0 = utils.rho_2(0, e)
    R1 = utils.rho_2(cm0, e)
    R2 = utils.rho_2(120, e)
    cm = x
    R = utils.rho_2(cm, e)
    if (180 < x) and (x < 360):
        cm = 360 - x
    if 0 <= cm <= cm0:
        s = s0 + (s1 - s0) * (R - R0) / (R1 - R0)
    else:
        s = s1 + (s2 - s1) * (R - R1) / (R2 - R1)
    return s


# # # # # # # # # # #
#       VENUS       #
# # # # # # # # # # #


@dmodel(Venus.lat_incl, 54, 288)
def venus_lat_incl(x, im):
    """
    component Venus latitude due to inclination
    :param x: nodal argument of latitude = angle between ascending node and epicycle center in degree
    :param im: maximum inclination of the deferent on the ecliptic in degree
    :return: inclination in degree
    """

    return utils.lat_inf_incl_0(x, im)


@dmodel(Venus.lat_deviation, 55, 296, 298)
def venus_lat_deviation(x, R, jm):
    """
    component Venus latitude due to  deviation at latitude nodal argument=0
    :param x: true argument in degree
    :param R: radius of the epicycle
    :param jm: maximum deviation of the epicycle
    :return: latitude deviation at latitude nodal argument = 0
    """
    return utils.lat_inf_devia_asc_node_0(x, jm, R)


@dmodel(Venus.lat_slant, 56, 291, 293, 294)
def venus_lat_slant_approximated(x, R, b3m, pm):
    """
    Venus latitude slant approximated due to the slant of the epicycle at the deferent apogee
    and according ptolemy CALCULATIONS
    :param x: true argument in degree
    :param R: radius of the epicycle
    :param b3m: maximum slant of the epicycle in degree
    :param pm: maximum value of equation of argument p(av, cm) in degree
    :return: slant at the deference apogee(top) in degree
    """

    return utils.lat_inf_slant_apo_def_0(x, b3m, pm, R)


@dmodel(Venus.lat_slant, 57, 291, 292, 293)
def venus_lat_slant_geometric(x, R, e, b3m):
    """
    Venus latitude slant geometric due to the slant of the epicycle at the deferent apogee
    and according to the Ptolemy THEORY
    :param x: true argument in degree
    :param R: radius of the epicycle
    :param e: eccentricity
    :param b3m: maximum deviation of the epicycle
    :return: venus latitude slant in degree
    """

    return utils.lat_inf_slant_apo_def_1(x, R, 60 + b3m, e)


@dmodel(Venus.lat_double_arg, 58, 301, 303, 304, 305, 306)
def venus_lat_double_arg(x, y, R, im, jm, b3m, pm):
    """
    Venus total latitude double argument
    :param x: true argument in degree
    :param y: nodal argument of latitude=angle between ascending node amd epicycle centre
    :param R: radius of the epicycle
    :param im: inclination of the deferent on the ecliptic in degree
    :param jm: maximum deviation on the epicycle in degree
    :param b3m: maximum latitude due to the slant of the epicycle in degree
    :param pm: maximum value of the equation of the argument
    :return:
    """
    beta_1 = utils.lat_inf_incl_0(y, im)
    beta_2 = utils.lat_inf_devia_asc_node_0(x, jm, R)
    beta_3 = utils.lat_inf_slant_apo_def_0(x, b3m, pm, R)

    return beta_1 + m.sin(RAD * (y + 90)) * beta_2 + m.sin(RAD * y) * beta_3


@dmodel(Venus.equ_center, 29, 90)
def venus_center_equ(x, e):
    """
    Venus center equation
    :param x: mean center in degree
    :param e: eccentricity
    :return: center equation in degree
    """
    return utils.q_2(x, e)


@dmodel(Venus.equ_anomaly_at_max_dist, 59, 95, 96)
def venus_equ_anomaly_at_the_max_dist(x, e, R):
    """
    Venus equation anomaly at the maximum distance
    :param x: true argument in degree
    :param e: eccentricity
    :param R: radius of the epicycle
    :return: equation anomaly in degree:p0(av)=p(av,0º)
    """
    return utils.planet_anomaly_0(x, R, 60 + e)


@dmodel(Venus.equ_anomaly_at_mean_dist, 60, 93)
def venus_equ_anomaly_at_mean_dist(x, R):
    """
    Venus equation anomaly at mean distance rho(cm)=60
    :param x: true argument in degree
    :param R: radius of the epicycle
    :return: equation anomaly in degree p(av, cm0)=p0(av)
    """

    return utils.planet_anomaly_0(x, R, 60)


@dmodel(Venus.equ_anomaly_at_min_dist, 61, 98, 99)
def venus_equ_anomaly_at_min_dist(x, e, R):
    """
    Venus equation anomaly at minimum distance, rho=60-e
    :param x: true argument in degree
    :param e: eccentricity
    :param R: radius of the epicycle
    :return: equation anomaly in degree, p2(av)=p(av,180º)
    """
    return utils.planet_anomaly_0(x, R, 60 - e)


@dmodel(Venus.equ_minuta_proportionalia, 62, 266, 489)
def venus_equ_proportional_minute(x, y, e, R):
    """
    Venus equation proportional minute
    :param x: true argument (av) in degree
    :param y: mean center (cm) in degree
    :param e: eccentricity
    :param R: radius of the epicycle
    :return: minutes proportional in degree
    """

    return utils.minuta_proportionalia(x, R, e, y)


@dmodel(Venus.planetary_stations, 71, 176, 177, 178, 179)
def first_stationary_point_of_venus_by_proportional_minutes(x, s0, s1, s2, e):
    """
    First stationary point of Venus by proportional minutes
    :param x: mean center in degree
    :param s0: first station true argument  at apogee  (cm=0) in degree
    :param s1: first station true argument at mean position (cm=cm0) in degree
    :param s2: first station true argument at perigee (cm=180)
    :param e: planetary eccentricity
    :return:
    """
    return utils.first_stationary_point_proport_minutes_0(x, e, s0, s1, s2)


@dmodel(Venus.planetary_stations, 72, 179, 180, 181)
def first_stationary_point_of_venus_by_calculation_and_proportional_minutes(x, e, R, vq):
    """
    First stationary point of Venus by calculation and proportional minutes
    :param x: mean center in degree
    :param e: planetary eccentricity
    :param R: planetary epicycle radius
    :param vq: velocity quotient
    :return: true argument of the venus first statin in degree
    """
    s0 = utils.first_stationary_point_all_planet_except_mercury_0(0, vq, R, e)
    cm0 = utils.mean_pos_epic_center_0(e)
    s1 = utils.first_stationary_point_all_planet_except_mercury_0(cm0, vq, R, e)
    s2 = utils.first_stationary_point_all_planet_except_mercury_0(180, vq, R, e)
    return utils.first_stationary_point_proport_minutes_0(x, e, s0, s1, s2)


# # # # # # # # # # #
#        MOON       #
# # # # # # # # # # #


@dmodel(Moon.equ_anomaly, 41, 54, 55)
def moon_anomaly_equ(x, y, e, R):
    """
    Moon anomaly equation : general formula
    :param x: true argument in degree
    :param y: center or double elongation
    :param e: lunar eccentricity
    :param R: lunar epicycle radius
    :return:
    """

    rho = utils.product_cosine_0(y, e) + m.sqrt(
        60 ** 2 - (utils.product_sine_0(y, e)) ** 2
    )
    return -utils.planet_anomaly_0(x, R, rho)


@dmodel(Moon.equ_center, 43, 52)
def moon_center_equ(x, e):
    """
    Moon center equation
    :param x: center or double elongation
    :param e: lunar eccentricity
    :return: moon center equation in degree
    """
    rho = utils.product_cosine_0(x, e) + m.sqrt(
        60 ** 2 - (utils.product_sine_0(x, e)) ** 2
    )
    return utils.planet_anomaly_0(x, e, rho)


@dmodel(Moon.lunar_lat, 46, 132)
def lat_of_the_moon(x, imax):
    """
    Latitude of the Moon
    :param x: argument of latitude in degree
    :param imax: maximum value
    :return: latitude of the moon in degree
    """

    return DEG * m.asin(m.sin(x * RAD) * m.sin(imax * RAD))


# # # # # # # # # # #
#        MARS       #
# # # # # # # # # # #


@dmodel(Mars.equ_center, 30, 79)
def mars_center_equ(x, e):
    """
    Mars center equation
    :param x:
    :param e: eccentricity
    :return:
    """

    return utils.q_2(x, e)


@dmodel(Mars.equ_anomaly_at_max_dist, 66, 84, 85)
def mars_equ_anomaly_at_max_dist(x, e, R):
    """
    Mars equation anomaly at maximum distance
    :param x: true argument in degree
    :param e: eccentricity
    :param R: radius of the epicycle
    :return:
    """
    return utils.planet_anomaly_0(x, R, 60 + e)


@dmodel(Mars.equ_anomaly_at_mean_dist, 68, 82)
def mars_equ_anomaly_at_mean_dist(x, R):
    """
    Mars equation anomaly at mean distance
    :param x: true argument in degree
    :param R: radius of the epicycle
    :return: equation in degree
    """
    return utils.planet_anomaly_0(x, R, 60)


@dmodel(Mars.equ_anomaly_at_min_dist, 69, 87, 88)
def mars_equ_anomaly_at_min_dist(x, e, R):
    """
    Mars equation anomaly at minimum distance
    :param x: true argument in degree
    :param e: eccentricity
    :param R: radius of the epicycle
    :return: equation in degree
    """

    return utils.planet_anomaly_0(x, R, 60 - e)


@dmodel(Mars.equ_minuta_proportionalia, 70, 311, 490)
def mars_equ_proportional_minutes(x, y, e, R):
    """
    Mars equation proportional minutes
    :param x: true argument in degree
    :param y: mean center in degree
    :param e: eccentricity
    :param R: radius of the epicycle
    :return:
    """

    return utils.minuta_proportionalia(x, R, e, y)


# # # # # # # # # # #
#      JUPITER      #
# # # # # # # # # # #


@dmodel(Jupiter.equ_center, 35, 68)
def jupiter_center_equ(x, e):
    """
    Jupiter center equation
    :param x:
    :param e: eccentricity
    :return:
    """
    return utils.q_2(x, e)


# # # # # # # # # # #
#       SATURN      #
# # # # # # # # # # #


# # # # # # # # # # #
# SPHERICAL ASTRONO #
# # # # # # # # # # #


@dmodel(SphericalAstronomical.meridian_altitude_of_the_sun, 28, 201, 202)
def meridian_altitude_of_the_sun(x: float, eps: float, phi: float) -> float:
    """
    Meridian altitude of the sun
    :param x:
    :param eps:
    :param phi:
    :return:
    """

    return 90 - phi + DEG * m.asin(m.sin(x * RAD) * m.sin(eps * RAD))


@dmodel(SphericalAstronomical.declination, 24, 3)
def declination(x: float, obl: float) -> float:
    """
    declination of the sun
    :param x: longitude of sun in degree
    :param obl: obliquity of the ecliptic in degree
    :return:
    """
    return utils.declin_0(x, obl)


@dmodel(SphericalAstronomical.ascensional_diff, 31, 19, 20)
def ascensional_diffs(x, eps, phi):
    """
    Ascenscional differences of a point on ecliptic
    :param x: longitude of this point in degree
    :param eps: inclination of ecliptic in degree
    :param phi: latitude of the observer in degree
    :return: ascensional difference in degree
    """

    return DEG * m.asin(m.tan(utils.declin_0(x, eps) * RAD) * m.tan(phi * RAD))


@dmodel(SphericalAstronomical.right_ascension, 48, 8)
def right_ascension(x, obl):
    """
    Right ascension
    :param x: longitude in degree
    :param obl: obliquity of the ecliptic
    :return: right ascension in degree
    """
    return utils.right_asc_0(x, obl)


@dmodel(SphericalAstronomical.oblique_ascension, 49, 12, 13, 14)
def oblique_ascension(x, oblra, oblad, phi):
    """
    Oblique ascension
    :param x:
    :param oblra: obliquity of the ecliptic (ra)
    :param oblad: obliquity of the ecliptic (ad)
    :param phi: geographical latitude
    :return:oblique ascension in degree
    """

    return utils.oblique_asc_0(x, oblra, oblad, phi)


@dmodel(SphericalAstronomical.length_of_daylight, 50, 193, 194, 195)
def length_daylight(x, oblra, oblad, phi):
    """
    Length of daylight
    :param x: longitude in degree
    :param oblra: obliquity of the ecliptic (ra) in degree
    :param oblad: obliquity of the ecliptic (ad) in degree
    :param phi: geographical latitude in degree
    :return: Length of the day in DEGREE, divide by 15 to have the result in equinoctial hours
    """
    return utils.oblique_asc_0(x + 180, oblra, oblad, phi) - utils.oblique_asc_0(
        x, oblra, oblad, phi
    )


# # # # # # # # # # #
#   EIGHTH SPHERE   #
# # # # # # # # # # #


# # # # # # # # # # #
#      ECLIPSE      #
# # # # # # # # # # #


# # # # # # # # # # #
#   MATHEMATICAL    #
# # # # # # # # # # #


@dmodel(Mathematical.sine, 26, 2)
def sine(x: float, R=160) -> float:
    """
    Model for product sinus
    :param x: first argument of sine
    :param R: radius of the circle
    :return: a float
    """
    return utils.product_sine_0(x, R)



"""
To create a new `Calendar` you need to subclass it, then instanciate it with an `Era`.
A `Calendar` subclass has to have a name, a list of `Month` and an intercalation
method to be valid.

>>> class NewCal(Calendar):
...     _name = "My New Calendar"
...     _months = [Month(31, 32, "FirstMonth"),
...                Month(20, 22, "SecondMonth"),
...                Month(50, name="ThirdMonth")]
...     def intercalation(self, year: int) -> bool:
...         return year % 7 == 0
>>> my_era = Era("MyEra", 1234)
>>> my_calendar = NewCal(my_era)
>>> my_date = Date(my_calendar, (26, 3, 42), 13.5)
>>> str(my_date)
'42 ThirdMonth 26 MyEra in My New Calendar 13:30'
>>> my_date.jdn
3851.0625
"""

import abc
from dataclasses import dataclass
from functools import cached_property, lru_cache
from typing import Callable, Dict, List, Optional, Tuple, Union

from astropy.time import Time

from kanon.units import BasedReal, Sexagesimal
from kanon.utils.types.number_types import Real

CALENDAR_REGISTRY: Dict[str, "Calendar"] = {}

__all__ = ("Julian", "Byzantine", "Arabic", "Persian", "Egyptian", "Month", "Era")


def hm_to_float(hours: int, minutes: int) -> float:
    """Convert time in hours and minutes to a fraction of a day

    :type hours: int
    :type minutes: int
    :rtype: float
    """
    if not (0 <= hours < 24 and 0 <= minutes < 60):
        raise ValueError("Incorrect time")
    return hours / 24 + minutes / 60 / 24


def float_to_hm(fraction: float) -> Tuple[int, int]:
    """Convert fraction of a day into hours and minutes

    :type fraction: float
    :rtype: float
    """
    if not (0 <= fraction < 1):
        raise ValueError("Incorrect time")

    time = fraction * 24

    hours = int(time)
    mins = int((time - int(time)) * 60)

    return hours, mins


def hours_to_day(hours: Real) -> float:
    """Convert number of hours into fraction of day

    :type fraction: float
    :rtype: float
    """
    return float(hours) / 24


@dataclass(frozen=True)
class Era:
    """
    Dataclass defining an era.

    >>> Era("A.D.", 1721424)
    Era(name='A.D.', epoch=1721424)

    :param name: Name of the era
    :param days_ly: Start of the era in Julian Day Number
    """

    name: str
    epoch: float

    def days_from_epoch(self, jdn: float) -> float:
        return jdn - self.epoch


@dataclass(frozen=True)
class Month:
    """
    Dataclass defining a `~kanon.calendars.Calendar`'s month.

    >>> Month(28, 29, 'Februarius', ['February'])
    Month(days_cy=28, days_ly=29, name='Februarius', variant=['February'])

    :param days_cy: Number of days in a common year
    :param days_ly: Number of days in a leap year, optional, defaults to `days_cy` value
    :param name: Name of the month, optional, defaults to ""
    :param variant: List of name variants of this month, optional
    """

    days_cy: int

    days_ly: Optional[int] = None
    name: str = ""
    variant: Optional[List[str]] = None

    def _clone_new_leap(self, new_leap):
        return Month(self.days_cy, new_leap, self.name)

    def days(self, leap=False) -> int:
        """Returns the month's number of days in common or leap year

        :param leap: Is it a leap year, defaults to False
        :rtype: int
        """
        return self.days_ly if leap and self.days_ly else self.days_cy


class Date:
    """
    Dataclass defining a date.

    >>> cal = Calendar.registry["Julian A.D."]
    >>> date = Date(cal, (1,2,3), 13)
    >>> str(date)
    '3 Februarius 1 A.D. in Julian 13:00'
    >>> date.jdn
    1721457.0416666667
    >>> str(date + 1)
    '4 Februarius 1 A.D. in Julian 13:00'
    """

    def __init__(
        self,
        calendar: "Calendar",
        ymd: Tuple[int, int, int],
        hours: Real = Sexagesimal(12),
    ):
        """
        :param calendar: Calendar used in this date.
        :type calendar: Calendar
        :param ymd: Year, month and days, expressed in the specified calendar.
        :type ymd: Tuple[int, int, int]
        :param hours: Number of hours, defaults to Sexagesimal(12);
        :type hours: Real, optional
        """
        if not 0 <= float(hours) < 24:
            raise ValueError("Time must be in the range [0;24,0[")

        self._calendar = calendar
        self._ymd = ymd
        self._hours = (
            hours.resize(1)
            if isinstance(hours, Sexagesimal)
            else Sexagesimal.from_float(float(hours), 2)
        )

        self._jdn = calendar.jdn_at_ymd(*ymd) - 0.5 + hours_to_day(hours)

    @property
    def calendar(self) -> "Calendar":
        """Calendar used in this date.

        :rtype: Calendar
        """
        return self._calendar

    @property
    def ymd(self) -> Tuple[int, int, int]:
        """Year, month and days, expressed in the specified calendar.

        :rtype: Tuple[int, int, int]
        """
        return self._ymd

    @property
    def hours(self) -> BasedReal:
        """Number of hours

        :rtype: Sexagesimal
        """
        return self._hours

    @property
    def jdn(self) -> float:
        """Date as a julian day number.

        :rtype: float
        """
        return self._jdn

    def to_calendar(self, cal: "Calendar") -> "Date":
        """Express this date in another calendar."""
        return cal.from_julian_days(self.jdn)

    def days_from_epoch(self) -> float:
        """Get number of days from the start of the calendar"""
        return self.jdn - self.calendar.era.epoch

    def to_time(self) -> Time:
        """Express this date as a `astropy.time.Time` object with ``jd`` format."""
        return Time(self.jdn, format="jd")

    def __add__(self, other: Union["Date", Real]) -> "Date":
        jdn: Real = other.jdn if isinstance(other, Date) else other
        return self.calendar.from_julian_days(self.jdn + jdn)

    def __sub__(self, other: Union["Date", Real]) -> "Date":
        jdn: Real = other.jdn if isinstance(other, Date) else other
        return self.calendar.from_julian_days(self.jdn - jdn)

    def __eq__(self, o: object) -> bool:
        return isinstance(o, Date) and self.jdn == o.jdn

    def __str__(self):
        year, month, days = self.ymd
        h, m = self.hours[0], self.hours[1]
        return (
            f"{days} {self.calendar.months[month-1].name} "
            f"{year} {self.calendar.era.name} in {self.calendar._name} "
            f"{'0' if h < 10 else ''}{h}:{'0' if m < 10 else ''}{m}"
        )


class Calendar(metaclass=abc.ABCMeta):
    """This abstract class defines calendar behaviors. You need to subclass this to
    create a working `Calendar`. You have to define its `interpolation` method, its `_name`,
    `_months` and maybe `_cycle`.
    """

    #: Registry of all calendars
    registry: Dict[str, "Calendar"] = CALENDAR_REGISTRY

    _name: str
    _months: List[Month]
    _era: Era
    _variant: str
    _cycle: Tuple[int, int] = (1, 0)

    def __new__(
        cls,
        era: Era,
        variant: str = "",
        months_mutation: Optional[Callable[[List[Month]], List[Month]]] = None,
    ) -> "Calendar":
        """
        :param era: Era used by this calendar.
        :type era: Era
        :param variant: Name of this variant, defaults to ""
        :type variant: str, optional
        :param months_mutation: Function transforming the Calendar class `months` list, defaults to None
        :type months_mutation: Optional[Callable[[List[Month]], List[Month]]], optional
        :raises ValueError: Raised when the calendar's name has already been used.
        """
        self: Calendar = super().__new__(cls)
        self._era = era

        self._variant = variant

        if self.name in cls.registry:
            raise ValueError(
                f"{self.name} already exists in the registry, you might want to"
                "specify a variant name"
            )
        cls.registry[self.name] = self

        self._months = (months_mutation or (lambda x: x))(self._months.copy())
        return self

    @property
    def name(self) -> str:
        """Name of this calendar

        :rtype: str
        """
        return f"{self._name} {self._era.name}" + (
            f" {self._variant}" if self._variant else ""
        )

    @property
    def months(self) -> List[Month]:
        """List of months

        :rtype: List[Month]
        """
        return self._months

    @property
    def cycle(self) -> Tuple[int, int]:
        """Cycle of common year and leap years (common, leap)

        :rtype: Tuple[int, int]
        """
        return self._cycle

    @property
    def era(self) -> Era:
        """Calendar era

        :rtype: Era
        """
        return self._era

    @cached_property
    def common_year(self) -> int:
        """Number of days in a common year

        :rtype: int
        """
        return sum(m.days_cy for m in self.months)

    @cached_property
    def leap_year(self) -> int:
        """Number of days in a leap year

        :rtype: int
        """
        return sum(m.days(True) for m in self.months)

    @cached_property
    def cycle_length(self) -> int:
        """Number of days in a leap cycle

        :rtype: int
        """
        return self.cycle[0] * self.common_year + self.cycle[1] * self.leap_year

    @abc.abstractmethod
    def intercalation(self, year: int) -> bool:
        """Is the specified year an intercalation year (leap)"""
        raise NotImplementedError

    @lru_cache
    def jdn_at_ymd(self, year: int, month: int, day: int) -> float:
        """Julian day number at the specified date in ymd"""

        is_leap = self.intercalation(year)

        if 0 > month or month > len(self.months):
            raise ValueError(
                f"The month entered ({month}) is invalid 1..{len(self.months)}"
            )
        mdn = self.months[month - 1].days(is_leap)
        if day > mdn or day < 1:
            raise ValueError(
                f"The day entered ({day}) is invalid in {self.months[month-1].name} 1..{mdn}"
            )
        if year == 0:
            raise ValueError("Year cannot be zero.")

        days = 0

        negative_year = year < 0
        _year_calc = year if negative_year else year - 1

        year_rem = abs(_year_calc) % sum(self.cycle)

        days += sum(
            self.leap_year if self.intercalation(y) else self.common_year
            for y in (
                range(year - year_rem, year)
                if not negative_year
                else range(year, year + year_rem)
            )
        )

        days += (abs(_year_calc) // sum(self.cycle)) * self.cycle_length

        if negative_year:
            days *= -1

        days += sum(m.days(is_leap) for m in self.months[: month - 1]) + day - 1

        return days + self.era.epoch

    def get_time(self, year: int, month: int, day: int) -> Time:
        """`astropy.time.Time` object at the specified date in ymd"""
        return Time(self.jdn_at_ymd(year, month, day), format="jd")

    @lru_cache
    def from_julian_days(self, jdn: float) -> Date:
        """Builds a `Date` object at the specified julian day number."""

        time = (jdn - 0.5) % 1

        jdn = round(jdn)

        year = (
            int(self.era.days_from_epoch(jdn) * sum(self.cycle) // self.cycle_length)
            + 1
        )
        if year < 1:
            year -= 1

        rem = jdn - self.jdn_at_ymd(year, 1, 1)
        for y in range(year, year + self.cycle_length):
            ylength = self.leap_year if self.intercalation(y) else self.common_year
            if rem >= ylength:
                rem -= ylength
                year += 1
            else:
                break

        leap = self.intercalation(year)
        for i, m in enumerate(self.months):
            ndays = m.days(leap)
            if rem < ndays:
                month = i + 1
                days = rem + 1
                break
            else:
                rem -= ndays

        return Date(self, (year, month, int(days)), Sexagesimal("24;0") * time)

    def __repr__(self) -> str:
        return f"Calendar({self.name})"


class Julian(Calendar):
    """
    Defines the Julian Calendar.
    https://en.wikipedia.org/wiki/Julian_calendar
    """

    _name = "Julian"
    _months = [
        Month(31, 31, "Ianuarius", ["January"]),
        Month(28, 29, "Februarius", ["February"]),
        Month(31, 31, "Martius", ["March"]),
        Month(30, 30, "Aprilis", ["April"]),
        Month(31, 31, "Maius", ["May"]),
        Month(30, 30, "Iunius", ["June"]),
        Month(31, 31, "Iulius", ["July"]),
        Month(31, 31, "Augustus", ["August"]),
        Month(30, 30, "September", ["September"]),
        Month(31, 31, "October", ["October"]),
        Month(30, 30, "November", ["November"]),
        Month(31, 31, "December", ["December"]),
    ]

    _cycle = (3, 1)

    def intercalation(self, year: int) -> bool:
        if year < 0:
            year += 1
        return year % 4 == 0


class Arabic(Calendar):
    """
    Defines the Arabic Calendar.
    https://en.wikipedia.org/wiki/Islamic_calendar
    """

    _name = "Arabic"
    _months = [
        Month(30, 30, "Muḥarram"),
        Month(29, 29, "Ṣafar"),
        Month(30, 30, "Rabīʿ al-awwal"),
        Month(29, 29, "Rabīʿ al-thānī"),
        Month(30, 30, "Jumādā l-ūlā"),
        Month(29, 29, "Jumādā l-thāniyya"),
        Month(30, 30, "Rajab"),
        Month(29, 29, "Shaʿbān"),
        Month(30, 30, "Ramaḍān"),
        Month(29, 29, "Shawwāl"),
        Month(30, 30, "Dhū l-qaʿda"),
        Month(29, 30, "Dhū l-ḥijja"),
    ]
    _cycle = (19, 11)

    def intercalation(self, year: int) -> bool:
        return (1 + (year + 29) % 30) in {2, 5, 7, 10, 13, 16, 18, 21, 24, 26, 29}


class Byzantine(Calendar):
    """
    Defines the Byzantine Calendar.
    https://en.wikipedia.org/wiki/Byzantine_calendar
    """

    _name = "Byzantine"
    _months = [
        Month(31, 31, "Adhār"),
        Month(30, 30, "Nisān"),
        Month(31, 31, "Ayyār"),
        Month(30, 30, "Ḥazirān"),
        Month(31, 31, "Tammūz"),
        Month(31, 31, "Āb"),
        Month(30, 30, "Aylūl"),
        Month(31, 31, "Tishrīn al-awwal"),
        Month(30, 30, "Tishrīn al-thānī"),
        Month(31, 31, "Kānūn al-awwal"),
        Month(31, 31, "Kānūn al-thānī"),
        Month(28, 29, "Shubāṭ"),
    ]

    _cycle = (3, 1)

    def intercalation(self, year: int) -> bool:
        return (year - 1) % 4 == 0


class Egyptian(Calendar):
    """
    Defines the Egyptian Calendar.
    https://en.wikipedia.org/wiki/Egyptian_calendar
    """

    _name = "Egyptian"
    _months = [
        Month(30, name="Thoth"),
        Month(30, name="Phaophi"),
        Month(30, name="Athyr"),
        Month(30, name="Choiak"),
        Month(30, name="Tybi"),
        Month(30, name="Mechir"),
        Month(30, name="Phamenoth"),
        Month(30, name="Pharmuthi"),
        Month(30, name="Pachon"),
        Month(30, name="Payni"),
        Month(30, name="Epiphi"),
        Month(30, name="Mesore"),
        Month(5, name="Epagomenai"),
    ]

    def intercalation(self, year: int) -> bool:
        return False


class Persian(Calendar):
    """
    Defines the Persian Calendar.
    https://en.wikipedia.org/wiki/Zoroastrian_calendar
    """

    _name = "Persian"
    _months = [
        Month(30, name="Farwardīn"),
        Month(30, name="Urdībihisht"),
        Month(30, name="Khurdādh"),
        Month(30, name="Tīr"),
        Month(30, name="Murdādh"),
        Month(30, name="Shahrīwar"),
        Month(30, name="Mihr"),
        Month(30, name="Ābān"),
        Month(30, name="Ādhar"),
        Month(30, name="Day"),
        Month(30, name="Bahman"),
        Month(30, name="Isfandārmudh"),
        Month(5, name="Andarjah"),
    ]

    def intercalation(self, year: int) -> bool:
        return False


# Arabic Calendars
Arabic(Era("Civil Hijra", 1948440))
Arabic(Era("Astronomical Hijra", 1948439))

# Egyptian Calendars
Egyptian(Era("Nabonassar", 1448638))
Egyptian(Era("Philippus", 1603398))

_anno_domini = Era("A.D.", 1721424)

# Julian Calendars
Julian(_anno_domini)


def _leap_december(months):
    months[1] = months[1]._clone_new_leap(28)
    months[-1] = months[-1]._clone_new_leap(32)
    return months


Julian(_anno_domini, variant="Leap December", months_mutation=_leap_december)
Julian(
    _anno_domini, variant="First month March", months_mutation=lambda m: m[2:] + m[:2]
)
Julian(Era("Julian Era", 0))

# Byzantine Calendars
Byzantine(_anno_domini)

# Persian Calendars
_yazdigird = Era("Yazdigird", 1952063)
Persian(_yazdigird, variant="Andarjah at the end")
Persian(
    _yazdigird,
    variant="Andarjah after Ābān",
    months_mutation=lambda m: m[:8] + [m[-1]] + m[8:-1],
)



from dataclasses import asdict
from decimal import Decimal

import pytest

from kanon.units import Sexagesimal
from kanon.units.precision import (
    PreciseNumber,
    PrecisionContext,
    PrecisionMode,
    TruncatureMode,
    _with_context_precision,
    clear_records,
    find_func,
    get_context,
    get_records,
    identify_func,
    remove_func,
    set_context,
    set_precision,
    set_recording,
)
from kanon.units.radices import BasedReal


class TestPrecision:
    @classmethod
    def setup_class(cls):
        cls.context = asdict(get_context())
        del cls.context["_records"]
        del cls.context["stack"]

    def test_context(self):

        current_ctx = get_context()
        ctx = PrecisionContext(1, TruncatureMode.CEIL, None, None, None, None)
        set_context(ctx)
        with set_precision() as ctx_dict:
            assert ctx_dict["pmode"] == 1
            with pytest.raises(ValueError):
                set_context(ctx)
        set_context(current_ctx)

        with pytest.raises(ValueError) as err:
            ctx.mutate(add=lambda x, y: 0)

        assert "not registered" in str(err)

    def equality(self, a: BasedReal, b: BasedReal):
        assert a.equals(
            b
        ), f"{a.truncate()},r:{a.remainder} != {b.truncate()},r:{b.remainder}"

    def test_precision_modes(self):
        s1 = Sexagesimal("0;30,0,0,6")
        s2 = Sexagesimal(2)
        with set_precision(pmode=PrecisionMode.MAX):
            self.equality(s1 + s2, Sexagesimal("2;30,0,0,6"))
            self.equality(s1 * s2, Sexagesimal("1;0,0,0,12"))
            self.equality(s1 / s2, Sexagesimal("0;15,0,0,3"))

        with set_precision(pmode=PrecisionMode.SCI):
            s1_ = s1.truncate(3)
            self.equality(s1_ + s2, Sexagesimal(2, remainder=Decimal("0.5")))
            self.equality(s1_ * s2, Sexagesimal(1))
            self.equality(s1_ / s2, Sexagesimal(0, remainder=Decimal("0.25")))

        with set_precision(pmode=3):
            self.equality(
                s1 + s2, Sexagesimal((2,), (30, 0, 0), remainder=Decimal("0.1"))
            )
            self.equality(
                s1 * s2, Sexagesimal((1,), (0, 0, 0), remainder=Decimal("0.2"))
            )
            self.equality(
                s1 / s2, Sexagesimal((0,), (15, 0, 0), remainder=Decimal("0.05"))
            )

            assert round(Sexagesimal(2, remainder=Decimal("0.5"))) == 3

        with pytest.raises(NotImplementedError):
            with set_precision(pmode=PrecisionMode.FULL):
                s1 + s2

        with pytest.raises(ValueError):
            with set_precision(pmode=-1):
                pass

        with pytest.raises(TypeError):
            with set_precision(pmode="a"):
                pass

    def test_truncature_modes(self):
        get_context().mutate(pmode=PrecisionMode.SCI)
        s1 = Sexagesimal("0;30,0,0,6")
        s2 = Sexagesimal(2)
        with set_precision(tmode=TruncatureMode.ROUND):
            self.equality(s1 + s2, Sexagesimal(3))
            self.equality(s1 * s2, Sexagesimal(1))
            self.equality(s1 / s2, Sexagesimal(0))
            self.equality(s1 / Sexagesimal(1), Sexagesimal(1))
        with set_precision(tmode=TruncatureMode.TRUNC):
            self.equality(s1 + s2, Sexagesimal(2))
            self.equality(s1 * s2, Sexagesimal(1))
            self.equality(s1 / s2, Sexagesimal(0))
            self.equality(s1 / Sexagesimal(1), Sexagesimal(0))
        with set_precision(tmode=TruncatureMode.FLOOR):
            self.equality(s1 + s2, Sexagesimal(2))
            self.equality(s1 * s2, Sexagesimal(1))
            self.equality(s1 / s2, Sexagesimal(0))
            self.equality(s1 / Sexagesimal(1), Sexagesimal(0))
            self.equality(-s1 + -s2, -Sexagesimal(3))
            self.equality(s1 * -s2, -Sexagesimal(2))
            self.equality(s1 / -s2, -Sexagesimal(1))
            self.equality(-s1 / Sexagesimal(1), -Sexagesimal(1))
        with set_precision(tmode=TruncatureMode.CEIL):
            self.equality(s1 + s2, Sexagesimal(3))
            self.equality(s1 * s2, Sexagesimal(2))
            self.equality(s1 / s2, Sexagesimal(1))
            self.equality(s1 / Sexagesimal(1), Sexagesimal(1))
            self.equality(-s1 + -s2, -Sexagesimal(2))
            self.equality(s1 * -s2, -Sexagesimal(1))
            self.equality(s1 / -s2, -Sexagesimal(0))
            self.equality(-s1 / Sexagesimal(1), -Sexagesimal(0))
        get_context().mutate(pmode=PrecisionMode.MAX)

        with pytest.raises(TypeError):
            with set_precision(tmode=1):
                pass

    def test_custom_arithmetic(self):
        @identify_func("ADD")
        def add(a: PreciseNumber, b: PreciseNumber):
            return a._add(Sexagesimal.from_float(float(b) + 1, 0))

        @identify_func("SUB")
        def sub(a: PreciseNumber, b: PreciseNumber):
            return a._sub(Sexagesimal.from_float(float(b) + 1, 0))

        @identify_func("MUL")
        def mul(a: PreciseNumber, b: PreciseNumber):
            return a._mul(Sexagesimal.from_float(float(b) + 1, 0))

        @identify_func("DIV")
        def div(a: PreciseNumber, b: PreciseNumber):
            return Sexagesimal(5)

        with set_precision(add=add, sub=sub, div=div, mul=mul):
            assert Sexagesimal(1) + Sexagesimal(1) == 3
            assert Sexagesimal(1) - Sexagesimal(1) == 0
            assert Sexagesimal(1) * Sexagesimal(1) == 2
            assert Sexagesimal(1) / Sexagesimal(1) == 5

        with pytest.raises(ValueError) as err:

            @identify_func("ADD")
            def _add():
                pass

        assert "already in use" in str(err)

        assert find_func("ADD") == add
        assert remove_func(add)
        assert not remove_func(add)

        with pytest.raises(ValueError) as err:
            identify_func("ADD")(mul)
        assert "Function" in str(err)

    def test_history(self):
        set_recording(True)
        with set_precision(pmode=5):
            Sexagesimal(1) + Sexagesimal(2)

        records = get_records()
        assert len(records) == 1
        rec = records[0]
        assert rec["args"] == (Sexagesimal(1), Sexagesimal(2), "+", Sexagesimal(3))
        assert rec["add"] == "DEFAULT"

        set_recording(False)
        with set_precision(pmode=5):
            Sexagesimal(1) + Sexagesimal(2)
        assert len(records) == 1

        set_recording(True)
        Sexagesimal(1) + Sexagesimal(1)
        assert len(records) == 2

        @identify_func("_ADD")
        def add(a, b):
            return Sexagesimal(0)

        with set_precision(add=add):
            Sexagesimal(1) + Sexagesimal(1)
        assert len(records) == 3
        assert records[-1]["add"] == "_ADD"

        with set_precision():
            with pytest.raises(ValueError):
                set_recording(False)

        set_recording(False)

        clear_records()

        assert len(records) == 0

    def test_wrapper(self):
        @_with_context_precision
        def func(a, b):
            return 1

        with pytest.raises(TypeError):
            func(Sexagesimal(1), Sexagesimal(2))
        assert func(1, 2) == 1

    @classmethod
    def teardown_class(cls):
        get_context().mutate(**cls.context)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

# Packages may add whatever they like to this file, but
# should keep this content at the top.
# ----------------------------------------------------------------------------

try:
    from .version import version as __version__
except ImportError:
    __version__ = ""
